/* funktion.c
 * erstellt:  Mon Nov 22 17:44:10 CET 1999
 * geaendert: Fre Jan 14 14:08:49 CET 2000
 * Autoren:  Mirko Belickin und Joachim Wagner
 */

#include "funktion.h"
#include <stdio.h>
#include <math.h>

#ifndef PI
  #define PI 3.14159265
#endif

double f1(double x);
double f2(double x);
double f3(double x);
double f4(double x);
double f5(double x);

double df1(double x);
double df2(double x);
double df3(double x);
double df4(double x);

funktionsbeispiel funktion[] = {
   {f1, df1, "e^(2x) - sin(x) - 2" },
   {f2, df2,  "2x - tan(x)" },
   {f3, df3,  "x^2 - 2" },
   {f4, df4,  "arctan(x)" },
   {f5, NULL, "floor []" }
};

const int funktion_anzahl = sizeof(funktion)/sizeof(funktion[0]);

void funktionenhilfe(FILE *datei, char *prefix)
{
  int i;
  for (i=0; i<funktion_anzahl; i++) 
    fprintf(datei,"%s%i: %s\n", prefix, i+1, funktion[i].name);
}

pfkt funktion_nr(int nr)
{
  return funktion[nr-1].wert;
}

/* funktion_ableitung(int nr) 
 * liefert einen Zeiger auf eine Funktion, die die
 * Ableitung fuer die angegebene Funktion berechnet.
 * Achtung: kann nicht mehrfach verwendet werden.
 */

static int num_abl_nr = -1;

double num_abl(double x);

pfkt funktion_ableitung(int nr)
{
  if (num_abl_nr >= 0) {
    fprintf(stderr,"Modul 'Funktion' nicht korrekt benutzt.\n");
    exit(1);
  }
  num_abl_nr = nr-1;
  if (0 == 0) /* verfahren == genau */
    return funktion[nr-1].abl;
  else
    return num_abl;
}

double num_abl(double x)
{
  fprintf(stderr,"Modul 'Funktion': num_abl noch nicht implementiert.\n");
  exit(1);
  return 0.0;
}


double f1(double x)
{
  return exp(2.0*x) - sin(x) - 2.0;
}

double df1(double x)
{
  return 2.0 * exp(2.0*x) - cos(x);
}


double f2(double x)
{
  return 2.0*x - tan(x);
}

double df2(double x)
{
  double h = tan(x);
  return 2.0 - (h*h+1.0);
}


double f3(double x)
{
  return x*x - 2.0;
}

double df3(double x)
{
  return 2.0*x;
}


double f4(double x)
{
  return atan(x);
}

double df4(double x)
{
  return 1.0 / (1.0 + x*x);
}

double f5(double x)
{
  return floor(x);
}
