/* muller.c
 * Bestimmt Nullstellen nach der Muller - Methode
 * erstellt: Die Jan 18 17:26:00 CET 2000
 * Autoren : Mirko Belickin und Joachim Wagner
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "funktion.h"
#include "namen.h"

extern funktionsbeispiel funktion[];
extern const int funktion_anzahl;

#define TEILEN_DURCH_NULL { \
      fprintf(out,"keine Nullstelle gefunden wegen Teilen durch Null\n"); \
      fprintf(out,"x%i = %lg, f(x%i) = %lg\n", \
                  i+2,x2,i+2,f(x2)); \
      fprintf(out,"(%lg)\n", x2); \
      break; }


/* Hauptprogramm 
 *
 */

int main(int argc, char *argv[])
{
  FILE *out = stdout;
  FILE *temp;
  int hilfe = 0;        /* 1 = Hilfe fuer Benutzer anzeigen */   
  int fehler = 0;       /* 1 = Programm falsch aufgerufen */
  int schalter = 1;     /* 0 = es folgen keine Schalter ('--' gefunden) */
  double x, h;
  double x0, x1, x2, a, b, c, h1, h2, d1, d2;
  int n, i, schritte;
  int schritteBeschraenkt;
  double (* f)(double);
  
  schritteBeschraenkt = 0;
  
  while ((schalter) && (argc>1) && (*argv[1]=='-')) {/* Schalter bearbeiten */
    switch ((argv[1])[1]) {
      case 'h': hilfe = 1;
                break;
      case 's': schritteBeschraenkt = 1;
                {
                  char *str = argv[1]+2;  /* fuer -sAnzahl */
                  if (!*str) {
                    if (argc>3) {         /* fuer -s Anzahl */
                      argv[1] = argv[0];
                      argv++; 
                      argc--;
                      str = argv[1];
                    }else{
                      fehler = 1;
                    }
                  }
                  if (!fehler) schritte = atoi(str);
                }
                break;
      case '-': schalter = 0;
                break;
      default:  fprintf(stderr,"Schalter %s unbekannt.\n", argv[1]);
                fehler = 1;
                break;
    }
    argv[1] = argv[0];
    argv++; 
    argc--;
  }
  
  if ((argc>6) || (argc<5)) fehler = 1;
  if ((hilfe) || (fehler)) {             /* Hilfetext anzeigen */
    fprintf(stderr,"\nBenutzung: "
                   "%s [Schalter] [--] x0 x1 x2 n [Ausgabedatei]\n\n",
                   argv[0]);
    fprintf(stderr,"Schalter: -h     fuer diese Hilfe\n");
    fprintf(stderr,"          -s n   beschraenkt die Iteration auf n\n");
    fprintf(stderr," x0 = erster erforderlicher Startwert\n");
    fprintf(stderr," x1 = zweiter erforderlicher Startwert\n");
    fprintf(stderr," x2 = dritter erforderlicher Startwert\n");
    fprintf(stderr," n = Nummer der Funktion\n");
    funktionenhilfe(stderr,"       ");
    fprintf(stderr,"\nMirko Belickin und Joachim Wagner - 18. Januar 2000\n");
    exit(fehler);
  }
  x0 = atof(argv[1]);
  x1 = atof(argv[2]);
  x2 = atof(argv[3]);
  n = atoi(argv[4]);
  if (n<1) {
    fprintf(stderr, "n muss mindestens 1 sein!\n");
    exit(1);
  }
  if (n>funktion_anzahl) {
    fprintf(stderr, "Funktion Nr. %i nicht definiert.\n", n);
    exit(1);
  }
  if (argc>5) {
    if (!(temp=fopen(argv[5],"wb"))) {
      fprintf(stderr,"Oeffnen der Datei %s fehlgeschlagen.", argv[5]);
      exit(1);
    }
    out = temp;
  }
  
  fprintf(out, "\nFunktion %s\n\n", funktion[n-1].name);

  f =funktion_nr(n);

  i=0;
  while ((!schritteBeschraenkt) || (i < schritte)) {
    h2 = x0 - x2;
    h1 = x1 - x2;
    c = h = f(x2);
    d2 = f(x0) - h;
    d1 = f(x1) - h;
    h = h2 * h1 * (h2 - h1);   /* Determinate des Gleichungssystems */
    if (fabs(h) < EPSILON)  
      TEILEN_DURCH_NULL;
    a = ((h1 * d2) - (h2 * d1)) / h;   /* Koeffizienten der Parabel */
    b = ((h2 * h2 * d1) - (h1 * h1 * d2)) / h;
    /* c = f(x2) siehe oben */
    if (fabs(a) < EPSILON) {   /* wie Sekantenmethode */
      if (fabs(b) < EPSILON)
        TEILEN_DURCH_NULL;
      x = x2 - c/b;      
    }else{                     /* Nullstelle der Parabel finden */
      h = b*b - 4*a*c; 
      if (h < 0) {
        fprintf(out,"keine Nullstelle gefunden wegen eines negativen "
                    "Radikanten\n");
        fprintf(out,"x%i = %lg, f(x%i) = %lg\n",
                    i+2,x2,i+2,f(x2));
        fprintf(out,"(%lg)\n", x2);
        break; 
      }  
      x = x2 - 0.5 * (b/a + sqrt(h)/a);
    }
    h = f(x);                  /* neues x auf Nullstelle pruefen */
    if (fabs(h) < EPSILON) {   /* Ja -> ausgeben und fertig */
      fprintf(out,"Nullstelle gefunden:\n");
      fprintf(out,"%lg\n", x);
      break; 
    }
    x0 = x1;                  
    x1 = x2;
    x2 = x;
    i++;                       /* Dreistufeniteration wiederholen */
  }
  if (schritteBeschraenkt && (i == schritte)) {
    fprintf(out,"Nach %i Schritten noch keine Nullstelle gefunden.\n",
                schritte); 
    fprintf(out,"x%i = %lg, f(x%i) = %lg\n",
                i+1,a,i+1,f(a));
    fprintf(out,"(%lg)\n", a);
  }
  
  if (out != stdout) fclose(out);               /* Dateien ggf. schliessen */
  return 0;
}
