/* regulafa.c
 * Bestimmt Nullstellen nach der Regula Falsi - Methode
 * erstellt: Fre Jan 14 14:25:56 CET 2000
 * Autoren : Mirko Belickin und Joachim Wagner
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "funktion.h"
#include "namen.h"

extern funktionsbeispiel funktion[];
extern const int funktion_anzahl;


/* Hauptprogramm 
 *
 */

int main(int argc, char *argv[])
{
  FILE *out = stdout;
  FILE *temp;
  int hilfe = 0;        /* 1 = Hilfe fuer Benutzer anzeigen */   
  int fehler = 0;       /* 1 = Programm falsch aufgerufen */
  int schalter = 1;     /* 0 = es folgen keine Schalter ('--' gefunden) */
  double a, b, x, h;
  int n, i, schritte;
  int schritteBeschraenkt;
  char tempname[80];
  double (* f)(double);
  
  schritteBeschraenkt = 0;
  
  while ((schalter) && (argc>1) && (*argv[1]=='-')) {/* Schalter bearbeiten */
    switch ((argv[1])[1]) {
      case 'h': hilfe = 1;
                break;
      case 's': schritteBeschraenkt = 1;
                {
                  char *str = argv[1]+2;  /* fuer -sAnzahl */
                  if (!*str) {
                    if (argc>3) {         /* fuer -s Anzahl */
                      argv[1] = argv[0];
                      argv++; 
                      argc--;
                      str = argv[1];
                    }else{
                      fehler = 1;
                    }
                  }
                  if (!fehler) schritte = atoi(str);
                }
                break;
      case '-': schalter = 0;
                break;
      default:  fprintf(stderr,"Schalter %s unbekannt.\n", argv[1]);
                fehler = 1;
                break;
    }
    argv[1] = argv[0];
    argv++; 
    argc--;
  }
  
  if ((argc>5) || (argc<4)) fehler = 1;
  if ((hilfe) || (fehler)) {             /* Hilfetext anzeigen */
    fprintf(stderr,"\nBenutzung: "
                   "%s [Schalter] [--] a b n [Ausgabedatei]\n\n",
                   argv[0]);
    fprintf(stderr,"Schalter: -h     fuer diese Hilfe\n");
    fprintf(stderr,"          -s n   beschraenkt die Iteration auf n\n");
    fprintf(stderr," a = linke Grenze des Startintervalls\n");
    fprintf(stderr," b = rechte Grenze des Startintervalls\n");
    fprintf(stderr," n = Nummer der Funktion\n");
    funktionenhilfe(stderr,"       ");
    fprintf(stderr,"\nMirko Belickin und Joachim Wagner - 14. Januar 2000\n");
    exit(fehler);
  }
  a = atof(argv[1]);
  b = atof(argv[2]);
  n = atoi(argv[3]);
  if (n<1) {
    fprintf(stderr, "n muss mindestens 1 sein!\n");
    exit(1);
  }
  if (n>funktion_anzahl) {
    fprintf(stderr, "Funktion Nr. %i nicht definiert.\n", n);
    exit(1);
  }
  if (argc>4) {
    if (!(temp=fopen(argv[4],"wb"))) {
      fprintf(stderr,"Oeffnen der Datei %s fehlgeschlagen.", argv[4]);
      exit(1);
    }
    out = temp;
  }
  
  fprintf(out, "\nFunktion %s\n\n", funktion[n-1].name);

  f=funktion_nr(n);

  if (f(a)*f(b)>=0) {
    fprintf(out,"Die Vorzeichen der Funktionswerte der Intervallgrenzen "
                "sind gleich. Es ist daher nicht sichergestellt, dass eine "
                "Nullstelle vorhanden ist.\n");
    if (schritteBeschraenkt) 
      fprintf(out,"Berechne trotzdem, Halt nach %i Schritten\n", schritte);
    else {
      fprintf(out,"Durch Angabe einer maximalen Anzahl von "
                  "Interationsschritten kann die Berechnung doch durchgefuehrt"
                  " werden.\n");
      exit(1);
    }
  }
  i=0;
  while ((!schritteBeschraenkt) || (i < schritte)) {
    h = f(b) - f(a);
    if (fabs(h) < EPSILON) {
      fprintf(out,"keine Nullstelle gefunden wegen Teilen durch Null\n");
      fprintf(out,"a = %lg, b = %lg, f(a) = %lg, f(b)=%lg\n",
                  a,b,f(a),f(b));
      fprintf(out,"(a+b)/2 = %lg, f((a+b)/2) = %lg\n",
                  (a+b)/2.0,f((a+b)/2.0));
      fprintf(out,"(%lg)\n", (a+b)/2.0);
      break; 
    }
    x = f(a) * (a-b) / h + a;
    h = f(x);
    if (fabs(h) < EPSILON) {
      fprintf(out,"Nullstelle gefunden:\n");
      fprintf(out,"%lg\n", x);
      break; 
    }
    if (h<0) 
      a = x;
    else
      b = x;
    i++;
  }
  if (schritteBeschraenkt && (i == schritte)) {
    fprintf(out,"Nach %i Schritten noch keine Nullstelle gefunden.\n",
                schritte); 
    fprintf(out,"a = %lg, b = %lg, f(a) = %lg, f(b)=%lg\n",
                a,b,f(a),f(b));
    fprintf(out,"(a+b)/2 = %lg, f((a+b)/2) = %lg\n",
                (a+b)/2.0,f((a+b)/2.0));
    fprintf(out,"(%lg)\n", (a+b)/2.0);
  }
  
  if (out != stdout) fclose(out);               /* Dateien ggf. schliessen */
  return 0;
}
