/* service.c
 * erstellt: Sat Apr  3 14:21:31 MEST 1999
 * Autoren:  Timo Steffens, Jahn-Takeshi Saito, Joachim Wagner
 */

#include "service.h"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

/* #define SERVICE_DEBUG */

#define JA 1
#define NEIN 0

unsigned int service_rand(service_randomseed *seed, unsigned int n)  
{
  unsigned int z;              
#ifndef SERVICE_RAND_OHNE_LCG
  int q, s;
#endif
#ifdef SERVICE_RAND_OHNE_AD
  z = 0;
#else                                  /* Generator Nr. 0: Add.Gen. */
  seed->ad[seed->adpos] += 
    seed->ad[(seed->adpos+SERVICE_RAND_ADCOEFF)%SERVICE_RAND_ADBREITE];
  z = seed->ad[seed->adpos];
  seed->adpos=(seed->adpos+1)%SERVICE_RAND_ADBREITE;
#endif
#ifndef SERVICE_RAND_OHNE_LCG
  /* Generatoren Nr. 1 und 2: LCGs (nach P. L'Ecuyer) */
#define MODULT(a,b,c,m,s) q=(s)/a;(s)=b*((s)-a*q)-c*q;if((s)<0)(s)+=m;
  MODULT(53668,40014,12211,2147483563,seed->s1)
  MODULT(52774,40692, 3791,2147483399,seed->s2)
  s = (seed->s1 - seed->s2); 
  if (s<1) s+= 2147483562;    /* Dreieckskurven sinnvoll ueberlagern */
  z ^= (unsigned)s;
#endif
  /* Rueckgabewert modulo n */
  if (n) {
    unsigned int rest;
    /* nur 2^32 - (2^32 mod n) benutzen !! */
    rest = ((unsigned)(0-n))%(unsigned)n;  /* (2^32 - n) mod n */
    rest = (unsigned)(-((int)rest));       /* (2^32 - rest)    */
    if (rest && (z>=rest)) {
#ifdef SERVICE_DEBUG
      fprintf(stderr,"[zusaetzlicher rand-Aufruf fuer n = %u da "
                     "z = %u (%u zu gross)]\n",n,z,(z-rest)+1);
#endif
      return service_rand(seed,n);
    }      
    else return z % n;
  }
  else return z;
}


double service_drand(service_randomseed *seed)
{
  double d = 0.0;
  d += (double)service_rand(seed, 1<<26);     /* 0 bis 2^26-1 */
  d /= (double)(1<<26);                       /* 0 bis 1      */
  d += (double)service_rand(seed, 1<<26);     /* 0 bis 2^26-1 */
  d /= (double)(1<<26);                       /* 0 bis 1      */
  return d;
}


void service_initseed(service_randomseed *seed, int a, int b, int c)
{
  int i;
  if (!b && !c) {
    c = a & 0x0003ff00;     /* 10 Bit:  8 - 17                   */
    b = a & 0x0ffc0000;     /* 10 Bit: 18 - 27                   */
    a = a & 0xf00000ff;     /* 12 Bit:  0 -  7, 28 - 31          */
  }
  if (b && !c) {
    c = (a & 0x33000000) +  /* 20 Bit: 24 - 25, 28 - 29,         */
        (b & 0x003fffc0);   /*          6 - 21                   */
    b = (a & 0x00f00000) +  /* 20 Bit: 20 - 23                   */
        (b & 0xffc0003f);   /*          0 -  9, 22 - 31          */
    a =  a & 0xcc0fffff;    /* 24 Bit:  0 - 19, 26 - 27, 30 - 31 */
  }
#ifdef SERVICE_RAND_OHNE_AD
  seed->ad[0] = 0;
  seed->s1 = b ^ (a & 0x55555555);  /* Bits 0,2,4,...,30 */
  seed->s2 = c ^ (a & 0xaaaaaaaa);  /* Bits 1,3,5,...,31 */
#else
#ifdef SERVICE_RAND_OHNE_LCG
  seed->ad[0] = (a^b)^c;
  seed->s1 = 0;
  seed->s2 = 0;
#else
  seed->ad[0] = a;
  seed->s1 = b;
  seed->s2 = c;
#endif
#endif
  for (i=1; i<SERVICE_RAND_ADBREITE; i++) {
    a = (1<<30)/SERVICE_RAND_ADBREITE;
    seed->ad[i] = i * a;
  }
  seed->adpos = 0;
  service_add2seed(seed,0);       /* Information gleichmaessig verteilen */
}

void service_zerstreuen(service_randomseed *seed);

void service_add2seed(service_randomseed *seed, int a)
{
  int t0, t1, t2;                  /* Hilfsvariablen fuer Ringtausch  */
  int i, j;
  for (j=0; j<32; j++) {
  i = seed->adpos;
  seed->ad[i] ^= (a & 0x50155415);
  seed->s1    ^= (a & 0xa82aa80a); 
  seed->s2    ^= (a & 0x07c003e0);
  service_zerstreuen(seed);        /* Werte zerstreuen         */
  t0 = seed->ad[i] & 0x00155400;   /* Ringtausch einiger Bits  */
  t1 = seed->s1    & 0x00155400;   /* Bits 10,12,...,20        */
  t2 = seed->s2    & 0x00155400;
  seed->ad[i] ^= (t0 ^ t1);        /* nach links               */
  seed->s1    ^= (t1 ^ t2);
  seed->s2    ^= (t2 ^ t0);
  service_zerstreuen(seed);        /* Werte zerstreuen         */
  t0 = seed->ad[i] & 0x002aa800;   /* Ringtausch einiger Bits  */
  t1 = seed->s1    & 0x002aa800;   /* Bits 11,13,...,21        */
  t2 = seed->s2    & 0x002aa800;
  seed->ad[i] ^= (t0 ^ t2);        /* nach rechts              */
  seed->s1    ^= (t1 ^ t0);
  seed->s2    ^= (t2 ^ t1);
  service_zerstreuen(seed);        /* Werte nochmal zerstreuen */
  }
}

void service_zerstreuen(service_randomseed *seed)
{
  int i, t;
  /* LCG-Werte muessen im richtigen Bereich liegen */
  if (!(seed->s1)) seed->s1++; 
  if ((seed->s1)>2147483562) seed->s1 &= 0x3fffffff;
  if (!(seed->s2)) seed->s2++; 
  if ((seed->s2)>2147483398) seed->s2 &= 0x3fffffff;
  /* Add. Gen. darf nicht null sein */
  for (i=t=0; i<SERVICE_RAND_ADBREITE; i++)
    if (seed->ad[i] & 1) t=1;        /* bilden LSBs ein LFSR?    */
  if (!t) seed->ad[seed->adpos] |= 1;
  for (i=0; i<SERVICE_RAND_ADBREITE; i++)
    service_rand(seed,0);            /* Werte nochmal zerstreuen */
}

int service_seedladen(FILE *datei, service_randomseed *seed)
{
  char eingabe[BUFSIZ];
  int i;
  /* ad */
  if (!fgets(eingabe,BUFSIZ,datei)) {
    fprintf(stderr,"Daten fuer Zufallszahlen-Keim (0) fehlen.\n");
    return NEIN; 
  }
  if (strncmp(eingabe,"randomseed0 = ",
               strlen("randomseed0 = "))!=0) {
    fprintf(stderr,"Daten fuer Zufallszahlen-Keim (0) sind ungueltig.\n");
    return NEIN; 
  }
  if (strlen(eingabe) < (strlen("randomseed0 = ") +
      SERVICE_RAND_ADBREITE* 9)) {
    fprintf(stderr,"Daten fuer Zufallszahlen-Keim unvollstaendig\n");
    return NEIN; 
  }
  for (i=0; i<SERVICE_RAND_ADBREITE; i++) { 
    if (sscanf(&eingabe[9*i+strlen("randomseed0 =")],
               " %x",&(seed->ad[i]))!=1) {  // unsigned
      fprintf(stderr,"Daten fuer Zufallszahlen-Keim (0) nicht numerisch.\n");
      return NEIN; 
    }
    seed->adpos = 0;
  }
  /* s1 */ 
  if (!fgets(eingabe,BUFSIZ,datei)) {
    fprintf(stderr,"Daten fuer Zufallszahlen-Keim (1) fehlen.\n");
    return NEIN; 
  }
  if (strncmp(eingabe,"randomseed1 = ",
               strlen("randomseed1 = "))!=0) {
    fprintf(stderr,"Daten fuer Zufallszahlen-Keim (1) sind ungueltig.\n");
    fprintf(stderr,"Inhalt: %s",eingabe);
    return NEIN; 
  }
  if (sscanf(eingabe,"randomseed1 = %d",&(seed->s1))!=1) {
    fprintf(stderr,"Daten fuer Zufallszahlen-Keim (1) nicht numerisch.\n");
    return NEIN; 
  }
  /* s2 */ 
  if (!fgets(eingabe,BUFSIZ,datei)) {
    fprintf(stderr,"Daten fuer Zufallszahlen-Keim (2) fehlen.\n");
    return NEIN; 
  }
  if (strncmp(eingabe,"randomseed2 = ",
               strlen("randomseed2 = "))!=0) {
    fprintf(stderr,"Daten fuer Zufallszahlen-Keim (2) sind ungueltig.\n");
    return NEIN; 
  }
  if (sscanf(eingabe,"randomseed2 = %d",&(seed->s2))!=1) {
    fprintf(stderr,"Daten fuer Zufallszahlen-Keim (2) nicht numerisch.\n");
    return NEIN; 
  }
  return JA;
}

int service_seedspeichern(FILE *datei, service_randomseed *seed)
{
  int i;
  fprintf(datei,"randomseed0 =");
  for (i=0; i<SERVICE_RAND_ADBREITE; i++)
    fprintf(datei," %08x",
            seed->ad[(i+seed->adpos)%SERVICE_RAND_ADBREITE]);
  fprintf(datei,"\n");
  fprintf(datei,"randomseed1 = %i\n",seed->s1);
  fprintf(datei,"randomseed2 = %i\n",seed->s2);
  return JA;
}


/*
 * *** Hash-Tabelle *** 
 */

static int geloeschtdummy;    /* Zeiger auf geloescht steht fuer GELOESCHT */
static void *geloescht;

/*
 * void service_inithash(service_hash *h, int groesse);    
 * richtet eine Hashtabelle ein
 */

void service_inithash(service_hash *h, int groesse)    
{
  int i;
  groesse += (groesse+8)/9;       /* + 11% */
  if (!(h->tabelle=(void **)malloc(groesse*sizeof(void *)))) {
    fprintf(stderr,"Nicht genug Speicher fuer Hilfstabelle"
           " (Hashing; %i)\n",groesse);
    exit(1);
  }
  for (i=0; i<groesse; i++) 
    h->tabelle[i] = NULL;
  h->groesse = groesse;
  h->schritte = 0;
  geloescht = (void *)&geloeschtdummy;
}


/* 
 * void service_freehash(service_hash *h);               
 * gibt Speicher frei
 */

void service_freehash(service_hash *h)               /* gibt Speicher frei */
{
  free(h->tabelle);
  h->tabelle = NULL;
  h->groesse = 0;
  h->schritte = 0;
}


/* 
 * int service_hashwert(service_hash *h, void *s)       
 * berechnet Hash-Wert fuer s 
 */

int service_hashwert(service_hash *h, void *s)
{
  service_randomseed seed;
  service_initseed(&seed,(int)s,0,0);        /* Zeigers als Zahl interpr. */
  return service_rand(&seed,h->groesse);
}


/* Kommentare aus GeHashing.java:
  // Unter Verwendung der Hashfunktion hash wird eine Hilfsmethode
  // implementiert, welche das Objekt in der Tabelle sucht und durch
  // - einen positiven Rueckgabewert den Index angibt, wo es gefunden wurde
  // - einen negativen Rueckgabewert den bitweise negierten Index angibt,
  //   wo es eingefuegt werden koennte (d.h., es wurde nicht gefunden).  */
/*   Implementiert ist ein Sondieren mittels zweitem Hashing.  */

int service_position(service_hash *h, void *s)  /* gibt Position der Daten  */
{
  int versuche          = 0;              /* Zahl der Sondierungen          */
  int d                 = 1;              /* Sondierschrittweite            */
  int loch_gefunden     = 0;              /* true, falls Loch gefunden      */
  int lochindex         = 0;              /* Vorschlag zum spaeteren Einf.  */
  int index   = service_hashwert(h,s);    /* berechne Hash-Wert             */
  service_randomseed seed;
  int max_versuche = 100 * h->groesse;

  while (!( (h->tabelle[index]==s)        /* Abbruch, falls Element gefunden*/
     || (h->tabelle[index]==NULL)         /* oder leeres Feld gefunden      */
     || (versuche == max_versuche))) {    /* oder Zahl der Vers. zu gross   */

                                          /* nur wegen spaeterem insert:    */
    if ((!loch_gefunden) &&               /* falls noch kein Loch gefunden  */
      (h->tabelle[index]== geloescht)) {  /* und GELOESCHT vorliegt,        */
      loch_gefunden = 1;                  /* merke, dass Loch gefunden      */
      lochindex     = index;              /* und seine Position             */
    }

    /*index = (index+d)%scheibenbenutzt;*//* mache einen Sondierschritt     */
    d    += 2;                            /* erhoehe Schrittweite           */
    service_initseed(&seed, index, d, (int)s);
    index = service_rand(&seed,h->groesse);
    versuche++;                           /* erhoehe Zahl der Versuche      */
  }

  if (versuche > (max_versuche/2)) {
    fprintf(stderr,"Warnung: %i Sondierschritte beim Hashing\n",
            versuche);
    fflush(stderr);
  }

  h->schritte = versuche;                 /* Anzahl veroeffentlichen */

  if (h->tabelle[index]==s)               /* falls belegtes Feld und s gef. */
    return index;                         /* liefere Position zurueck       */
  if (loch_gefunden) return ~lochindex;   /* liefere Loch (bitweise negiert)*/
                  else return ~index;     /* sonst letzte Einfuegeposition  */
}


/*
 * int service_einfuegen(service_hash *h, void *s) 
 * versucht, s einzufuegen       
 */

int service_einfuegen(service_hash *h, void *s) 
{
  int pos = service_position(h,s);        /* versuche, s zu finden          */
  if ((pos < 0) &&                        /* falls s nicht gefunden         */
      ((h->tabelle[~pos]==NULL) ||        /* und falls Platz vorhanden:     */
       (h->tabelle[~pos]== geloescht))){    
    h->tabelle[~pos] = s;                 /* fuege x ein / Zustand BELEGT   */
    return 1;                             /* melde Erfolg                   */
  }else{
    fprintf(stderr,"\nObjekt @%p konnte nicht in die Hilfstabelle"
                   " (Hashing) eingefuegt werden\n", s);
    /* exit(1); */
  }
  return 0;                               /* melde Misserfolg               */
}


/*
 * int service_nachschlag(service_hash *h, void *s) 
 * versucht, s nachzuschlagen 
 */

int service_nachschlag(service_hash *h, void *s) 
{
  int pos = service_position(h,s);        /* versuche, x zu finden          */
  if (pos >= 0) return pos;               /* falls gefunden: liefere Pos.,  */
           else return -1;                /* sonst melde Misserfolg         */
}


/*
 * int service_loeschen(service_hash *h, void *s);     
 * versucht, s zu loeschen    
 */

int service_loeschen(service_hash *h, void *s)  
{
  int pos = service_position(h,s);        /* versuche, s zu finden          */
  if (pos >= 0) {                         /* falls s gefunden:              */
    h->tabelle[pos] = geloescht;          /* setze Zustand auf GELOESCHT    */
    return 1;                             /* melde Erfolg                   */
  } else return 0;                        /* melde Misserfolg               */
}

/*
 * Speicherverwaltung
 */

typedef struct service_poolitem {
    struct service_poolitem *prec,
                            *succ;
    int daten;           /* ertser Block: Groesse des groessten Blocks */
                        /* sonst: Groesse des Blocks */
    void *malloc;
} service_poolitem;

#define SERVICE_UNGUELTIG 0
#define SERVICE_GUELTIG   1
#define SERVICE_MALLOC    2
#define SERVICE_PSCHRITTE 16

void service_initmpool(service_mpool *pool, unsigned int poolsize, int extra)
{
  int i;
  poolsize += 2*sizeof(service_poolitem);
  pool->pool_size = (poolsize+sizeof(service_poolitem)-1)
                       / sizeof(service_poolitem);
  if (!(pool->pool=(service_poolitem**)
        malloc(SERVICE_PSCHRITTE*sizeof(service_poolitem*)))) 
  {
    fprintf(stderr,"Nicht genug Speicher fuer die Speicherver"
                   "waltung (service_mpoolinit)\n");
    exit(1);
  }   
  for (i=0; i<SERVICE_PSCHRITTE; i++)
    pool->pool[i] = NULL;
  pool->free            =  0;
  pool->pos_free        =  0;
  pool->pos_first       =  0;
  pool->calls_2_alloc   =  0;
  pool->calls_2_dealloc =  0;
  pool->calls_2_malloc  =  1;
  pool->calls_2_free    =  0;
  /*
   * Hashtabelle fuer mit malloc belegte Flaechen 
   */
  service_inithash(&pool->hash, extra);
  pool->mhash = (service_poolitem **) 
                malloc(pool->hash.groesse*sizeof(service_poolitem *));
  if (!pool->mhash) {
    fprintf(stderr,"\nWarnung: Speicherverwaltung kann extra-grosse "
                   "Speicherflaechen nicht bereitsstellen.\n");
  }else{
    pool->calls_2_malloc++;
  }
}

void service_deallocall(service_mpool *pool)
{
  service_poolitem **pp;
  service_poolitem *p;
  pp = pool->pool;
  while (*pp) {
    p = *pp;
    do {
      if (p->malloc) {
        free(p->malloc);
        p->malloc = NULL;
        pool->calls_2_free++;
      }
    } while ((p=p->succ));   
    free(*pp);
    pool->calls_2_free++;
    pp++;
  }  
  service_freehash(&pool->hash);
  free(pool->mhash);
  pool->calls_2_free++;
  if (pool->calls_2_free+1 != pool->calls_2_malloc) {
    fprintf(stderr,"\nWarnung (Speicherverwaltung): "
                   "calls_2_malloc = %i\n"
                   "calls_2_free   = %i\n",
                   pool->calls_2_malloc, pool->calls_2_free+1);
  }
  free(pool->pool);
  /* pool->calls_2_free++; */
}

service_poolitem *service_item4(service_mpool *pool, unsigned int size);

void *service_alloc(service_mpool *pool, unsigned int size)
{
  void *p;
  service_poolitem *item;
  int bedarf, hpos;
  bedarf = (size+sizeof(service_poolitem)-1)
              / sizeof(service_poolitem);
  if (bedarf >= (pool->pool_size-1))
  {
    if (!(p=malloc(size))) return NULL;
    pool->calls_2_malloc++;
    service_einfuegen(&pool->hash, p);
    hpos = service_nachschlag(&pool->hash, p);
    if (hpos >= 0) {
      item = service_item4(pool, 0);
      if (item) {
        item->malloc = p;
        pool->mhash[hpos] = item;
        return p;
      }
      service_loeschen(&pool->hash, p);
    }
    free(p);
    pool->calls_2_free++;
    return NULL;
  }else{
    item = service_item4(pool, bedarf);
    if (!item) return NULL;
    p = item+1;
    return p;
  }
}

void service_calc_free(service_mpool *pool, service_poolitem *item);

void service_dealloc(service_mpool *pool, void *ptr)
{
  service_poolitem *item;
  int hpos;
  hpos = service_nachschlag(&pool->hash, ptr);
  if (hpos >= 0) {
    item = pool->mhash[hpos];
    if (item->malloc != ptr) {
      fprintf(stderr,"\nFehler: Speicherverwaltung konnte die Speicher"
                     "flaeche @%p nicht eindeutig identifizieren.\n", ptr);
      return; 
    }
    free(ptr);
    pool->calls_2_free++;
    item->malloc = NULL;
    service_loeschen(&pool->hash, ptr);
  }else{
    item = (service_poolitem*)ptr;
    item--;
  }
  item->prec->succ = item->succ;  /* aus der Verkettung nehmen */
  if (item->succ)
    item->succ->prec = item->prec; 
  service_calc_free(pool, item->prec);
}

service_poolitem *service_item4(service_mpool *pool, unsigned int bloecke)
{
  unsigned int bedarf;
  service_poolitem *p, *prec, *succ;
  int pos, luecke;
  bedarf = 1 + bloecke;
  if (pool->free >= bedarf) {
    pos = pool->pos_free; 
  }else{
    pos = pool->pos_first - 1;
    do {
      pos++;
      if (!(pool->pool[pos])) {      /* Ende der Liste erreicht? */
        if (!((pos+1) & (SERVICE_PSCHRITTE-1))) { /* realloc noetig? */
          service_poolitem **neu; 
          int i;
          if (!(neu=(service_poolitem**)
                realloc(pool->pool, (pos+1+SERVICE_PSCHRITTE) 
                                    * sizeof(service_poolitem*)))) 
          {
            fprintf(stderr,"Vergroesserung der verwalteten Speicher"
                           "flaeche fehlgeschlagen\n");
            return NULL;
          }
          pool->pool = neu;
          for (i=(pos+1); i<(pos+1+SERVICE_PSCHRITTE); i++)
            pool->pool[i] = NULL;
        }
        if (!(pool->pool[pos]=(service_poolitem*)
            malloc(sizeof(service_poolitem)*pool->pool_size))) 
          return NULL;
        pool->calls_2_malloc++;
        pool->pool[pos]->prec  = NULL;
        pool->pool[pos]->succ  = NULL;
        pool->pool[pos]->daten = pool->pool_size - 1;
        pool->pool[pos]->malloc = 0;
        if (pool->free < (pool->pool_size-1)) {
          pool->free = pool->pool_size - 1;
          pool->pos_free = pos;
        }
      }
    } while (pool->pool[pos]->daten < bedarf);
  }
  /* Nun gibt pos einen geeigneten Block an. */
  p = pool->pool[pos];
  luecke = 0;
  while (p->succ) {
    luecke = (p->succ - p - 1);
    if (p!=pool->pool[pos])
      luecke -= p->daten; 
    if (luecke >= bedarf) break;
    p = p->succ;
  }
  if (!(p->succ)) {      
    luecke = ((&pool->pool[pos][pool->pool_size]) - p - 1);
    if (p!=pool->pool[pos])
      luecke -= p->daten;
  }
  if (luecke >= bedarf) {   
    prec = p;
    succ = p->succ;
    if (p == pool->pool[pos]) {
      p++;
    }else{
      p = p + p->daten + 1;
    }
    p->prec   = prec;
    p->succ   = succ;
    p->daten  = bedarf - 1;
    p->malloc = NULL;
    p->prec->succ = p;
    if (p->succ) 
      p->succ->prec = p;
    service_calc_free(pool, p);
  }else{  
    /* Block war doch nicht geeignet */
    fprintf(stderr,"Warnung: Inkonsistenz in der Speicherverwaltung "
                   "(service_item4) gefunden.\n");
    fprintf(stderr,"Versuche, sie zu beheben ...\n");
    fflush(stderr);
    service_calc_free(pool, pool->pool[0]); /* Neuberechnung */
    p = service_item4(pool, bloecke);  
    fprintf(stderr,"... Inkonsistenz in der Speicherverwaltung behoben\n");
    fflush(stderr);
  }
  return p;
}

void service_calc_free(service_mpool *pool, service_poolitem *item)
{
  service_poolitem *p, *start, *hilf;
  int pos, luecke, maxluecke;
  start = NULL;
  p = item;
  maxluecke = -1;
  do {
    if (p->succ) {
      luecke = (p->succ - p - 1);
    }else{
      if (!start) {
        start = p;
        while (start->prec)
          start = start->prec;
      }
      luecke = ((&start[pool->pool_size]) - p - 1);
    }
    if (p->prec)
      luecke -= p->daten;
    if (luecke >= maxluecke)
      maxluecke = luecke; 
    hilf = p;
    p = p->prec;
  } while (p); 
  start = hilf;
  p = item;
  do {
    if (p->succ) {
      luecke = (p->succ - p - 1);
    }else{
      luecke = ((&start[pool->pool_size]) - p - 1);
    }
    if (p->prec)
      luecke -= p->daten;
    if (luecke >= maxluecke)
      maxluecke = luecke; 
    p = p->succ;
  } while (p); 
  if (maxluecke < 0) {
    fprintf(stderr,"Fehler: maxluecke < 0 in service_calc_free\n");
    exit(1);
  }
  start->daten = maxluecke;
  pos = pool->pos_first;
  pool->pos_free  = 0;
  pool->free      = 0;
  while ((p = pool->pool[pos])) {
    if (pool->free < p->daten) {
      pool->free = p->daten;
      pool->pos_free = pos;
    } 
    if (   (pos>0) 
        && (pool->pos_first==(pos-1))
        && (!pool->pool[pos-1]->daten))
      pool->pos_first=pos;
    pos++;
  }
}
