/* service.h
 * erstellt: Sat Apr  3 14:21:31 MEST 1999
 * Autoren:  Timo Steffens, Jahn-Takeshi Saito, Joachim Wagner
 */

#ifndef __SERVICE_H
#define __SERVICE_H

#include <stdio.h>

/*
 * Zufallszahlen
 */

/* #define SERVICE_RAND_OHNE_AD */
/* #define SERVICE_RAND_OHNE_LCG */

#define SERVICE_RAND_ADBREITE 20
#define SERVICE_RAND_ADCOEFF (SERVICE_RAND_ADBREITE-3)

/* alt. Polynom (20,3,0)
 * akt. Polynom (7,3,0) 
 */

typedef struct service_randomseed {
    unsigned int ad[SERVICE_RAND_ADBREITE];  /* Additiver Generator */
    int adpos;
    int s1, s2;                              /* zwei LCGs */
} service_randomseed;

unsigned int service_rand(service_randomseed *seed, unsigned int n);  
/* liefert Zufallszahl von 0 - n-1, n==0 -> 32-Bit Zahl             */
/* n maximal 2^31-1                                                 */
/* AD-Periode: >=max(2^ADBREITE-1, 2^32-1)  LCG-Periode: ca. 10^18  */
/* Achtung: Der LCG liefert nicht alle 2^32 Zahlen und hat eine     */
/* nicht konstante Verteilungskurve.                                */

double service_drand(service_randomseed *seed);

void service_initseed(service_randomseed *seed, int a, int b, int c);
/* falls b und c oder nur c null sind, wird  */
/* a bzw werden a und b entsprechen verteilt */

void service_add2seed(service_randomseed *seed, int a);
int service_seedladen(FILE *datei, service_randomseed *seed);
int service_seedspeichern(FILE *datei, service_randomseed *seed);


/*
 * Hash-Tabelle
 */

typedef struct service_hash {
    void ** tabelle;
    int groesse;
    int schritte;        /* zuletzt benoetigte Sondierschritte, nur Info */
} service_hash;

void service_inithash(service_hash *h, int groesse);  /* Tabelle einrichten */
/* groesse wird von inithash erhoeht, damit weniger Schritte noetig sind */
void service_freehash(service_hash *h);               /* gibt Speicher frei */
int service_hashwert(service_hash *h, void *s);       /* Hash-Wert fuer s   */
int service_position(service_hash *h, void *s);       /* Position der Daten */
int service_einfuegen(service_hash *h, void *s);      /* s einfuegen        */
int service_nachschlag(service_hash *h, void *s);     /* s nachschlagen     */
int service_loeschen(service_hash *h, void *s);       /* s loeschen         */

/* vergleiche GeHashing.java aus ainf */


/*
 * Speicherverwaltung
 */


struct service_poolitem;

typedef struct service_mpool {
    struct service_poolitem **pool;
    int pool_size;
    int free;           /* Groesse des groessten Bloecke */
    int pos_free;       /* Stelle des groessten Bloecke */
    int pos_first;      /* Stelle des ersten groesseren Blocks */
    service_hash hash;
    struct service_poolitem **mhash;
    int calls_2_alloc;  /* Statistik */
    int calls_2_dealloc;
    int calls_2_malloc;
    int calls_2_free;
} service_mpool;

void service_initmpool(service_mpool *pool, unsigned int size, int extra);
void service_deallocall(service_mpool *pool);
void *service_alloc(service_mpool *pool, unsigned int size);
void service_dealloc(service_mpool *pool, void *ptr);

#endif
