/* vereinfa.c
 * Bestimmt Nullstellen nach der vereinfachten Newton - Methode
 * erstellt: Die Jan 18 17:22:30 CET 2000
 * Autoren : Mirko Belickin und Joachim Wagner
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "funktion.h"
#include "namen.h"

extern funktionsbeispiel funktion[];
extern const int funktion_anzahl;


/* Hauptprogramm 
 *
 */

int main(int argc, char *argv[])
{
  FILE *out = stdout;
  FILE *temp;
  int hilfe = 0;        /* 1 = Hilfe fuer Benutzer anzeigen */   
  int fehler = 0;       /* 1 = Programm falsch aufgerufen */
  int schalter = 1;     /* 0 = es folgen keine Schalter ('--' gefunden) */
  double a, b, x, h;
  int n, i, schritte;
  int schritteBeschraenkt;
  double (* f)(double);
  double (* df)(double);
  double dfa;
  
  schritteBeschraenkt = 0;
  
  while ((schalter) && (argc>1) && (*argv[1]=='-')) {/* Schalter bearbeiten */
    switch ((argv[1])[1]) {
      case 'h': hilfe = 1;
                break;
      case 's': schritteBeschraenkt = 1;
                {
                  char *str = argv[1]+2;  /* fuer -sAnzahl */
                  if (!*str) {
                    if (argc>3) {         /* fuer -s Anzahl */
                      argv[1] = argv[0];
                      argv++; 
                      argc--;
                      str = argv[1];
                    }else{
                      fehler = 1;
                    }
                  }
                  if (!fehler) schritte = atoi(str);
                }
                break;
      case '-': schalter = 0;
                break;
      default:  fprintf(stderr,"Schalter %s unbekannt.\n", argv[1]);
                fehler = 1;
                break;
    }
    argv[1] = argv[0];
    argv++; 
    argc--;
  }
  
  if ((argc>4) || (argc<3)) fehler = 1;
  if ((hilfe) || (fehler)) {             /* Hilfetext anzeigen */
    fprintf(stderr,"\nBenutzung: "
                   "%s [Schalter] [--] x0 n [Ausgabedatei]\n\n",
                   argv[0]);
    fprintf(stderr,"Schalter: -h     fuer diese Hilfe\n");
    fprintf(stderr,"          -s n   beschraenkt die Iteration auf n\n");
    fprintf(stderr," x0 = erforderlicher Startwert\n");
    fprintf(stderr," n = Nummer der Funktion\n");
    funktionenhilfe(stderr,"       ");
    fprintf(stderr,"\nMirko Belickin und Joachim Wagner - 18. Januar 2000\n");
    exit(fehler);
  }
  a = atof(argv[1]);
  n = atoi(argv[2]);
  if (n<1) {
    fprintf(stderr, "n muss mindestens 1 sein!\n");
    exit(1);
  }
  if (n>funktion_anzahl) {
    fprintf(stderr, "Funktion Nr. %i nicht definiert.\n", n);
    exit(1);
  }
  if (argc>3) {
    if (!(temp=fopen(argv[3],"wb"))) {
      fprintf(stderr,"Oeffnen der Datei %s fehlgeschlagen.", argv[3]);
      exit(1);
    }
    out = temp;
  }
  
  fprintf(out, "\nFunktion %s\n\n", funktion[n-1].name);

  f =funktion_nr(n);
  df=funktion_ableitung(n);

  dfa=df(a);

  i=0;
  while ((!schritteBeschraenkt) || (i < schritte)) {
    h = dfa;
    if (fabs(h) < EPSILON) {
      fprintf(out,"keine Nullstelle gefunden wegen Teilen durch Null\n");
      fprintf(out,"x%i = %lg, f(x%i) = %lg\n",
                  i+1,a,i+1,f(a));
      fprintf(out,"(%lg)\n", a);
      break; 
    }
    x = a - f(a) / h;
    h = f(x);
    if (fabs(h) < EPSILON) {
      fprintf(out,"Nullstelle gefunden:\n");
      fprintf(out,"%lg\n", x);
      break; 
    }
    a = x;
    i++;
  }
  if (schritteBeschraenkt && (i == schritte)) {
    fprintf(out,"Nach %i Schritten noch keine Nullstelle gefunden.\n",
                schritte); 
    fprintf(out,"x%i = %lg, f(x%i) = %lg\n",
                i+1,a,i+1,f(a));
    fprintf(out,"(%lg)\n", a);
  }
  
  if (out != stdout) fclose(out);               /* Dateien ggf. schliessen */
  return 0;
}
